#define STRICT
#include <windows.h>
#include <stdio.h>
#include <math.h>
#include "orbitersdk.h"
#include "mfd.h"
#include "TransXFunction.h"
#include "transx.h"



TransXFunction::TransXFunction(class transxstate *tstate, OBJHANDLE thmajor, OBJHANDLE thminor, OBJHANDLE thtarget, OBJHANDLE thcraft, OBJHANDLE thbase)
{	
	state=tstate;
	sethandles(thmajor,thminor,thtarget,thcraft,thbase);
	simstartMJD=oapiTime2MJD(0);
	initpens();
}

TransXFunction::TransXFunction(class transxstate *tstate, OBJHANDLE thmajor, OBJHANDLE thminor)
{
	hmajor=thmajor;
	hminor=thminor;
	state=tstate;
	hmajtarget=NULL;
	hbase=NULL;
	hcraft=oapiGetFocusObject();
	simstartMJD=oapiTime2MJD(0);
	initpens();
}



TransXFunction::~TransXFunction()
{
	//Need to delete the pens to stop them eating system resources
	deletepens();
}

bool initialisevars() //Virtual function overloaded by other functions
{
	return true;
}  

void TransXFunction::saveself(FILEHANDLE scn)
{
	//Write terminator
	oapiWriteScenario_string(scn,"Finish","TransXFunction");
} //Should never be used - overloaded by further functions

void TransXFunction::restoreself(FILEHANDLE scn)
{
	//Search for the terminator, then return
	char *bufferpointer;
	char tempbuffer[18], finalbuffer[18];
	bool ok;
	do {
		ok=oapiReadScenario_nextline(scn,bufferpointer);
		strncpy(tempbuffer,bufferpointer,16);
		sscanf(tempbuffer,"%s",finalbuffer);
	}
	while (strcmp(finalbuffer,"Finish")!=0 && ok==true);
}

bool TransXFunction::loadhandle(FILEHANDLE scn,OBJHANDLE *handle)
{
	char *member,*bufferpointer;
	int length;
	bool ok=oapiReadScenario_nextline(scn,bufferpointer);
	if (!ok) return false;//Unexpected end of file found
	parser.parseline(bufferpointer);
	if (!parser.getlineelement(0,&member,&length)) return false;
	if (strcmp(member,"Finish")==0) return false;
	if (strcmp(member,"Handle")!=0) return true;
	if (!parser.getlineelement(1,&member,&length)) return false;//Field missing
	if (strcmp(member,"NULL")==0)
	{
		*handle=NULL;
		return true;//Handle set null
	}
	OBJHANDLE thandle=oapiGetObjectByName(member);
	if (thandle!=NULL) *handle=thandle;
	return true;
}


void TransXFunction::findfinish(FILEHANDLE scn)
{
	char *tbuffer,*member;
	int length;
	do
	{
		if (!oapiReadScenario_nextline(scn,tbuffer)) return;
		parser.parseline(tbuffer);
		parser.getlineelement(0,&member, &length);
	}
	while (strcmp(member,"Finish"));
}


void TransXFunction::savedouble(FILEHANDLE scn, double savenumber)
{
	char buffer[80];
	sprintf(buffer," %.12g",savenumber);
	oapiWriteScenario_string(scn,"Double",buffer);
}

void TransXFunction::savevector(FILEHANDLE scn, VECTOR3 &vector)
{
	char buffer[100];
	sprintf(buffer," %.12g %.12g %.12g",vector.x,vector.y,vector.z);
	oapiWriteScenario_string(scn,"Vector",buffer);
}


bool TransXFunction::loaddouble(FILEHANDLE scn, double *loadednumber)
{
	char *member,*bufferpointer;
	int length;
	bool ok=oapiReadScenario_nextline(scn,bufferpointer);
	if (!ok) return false;//Unexpected end of file found
	parser.parseline(bufferpointer);
	if (!parser.getlineelement(0,&member,&length)) return false;
	if (strcmp(member,"Finish")==0) return false; //Unexpected end of function found
	if (strcmp(member,"Double")!=0) return true; //Not a double
	if (!parser.getlineelement(1,&member,&length)) return false;
	*loadednumber=atof(member);
	return true;
}

bool TransXFunction::loadint(FILEHANDLE scn, int *loadedint)
{
	char *member,*bufferpointer;
	int length;
	bool ok=oapiReadScenario_nextline(scn,bufferpointer);
	if (!ok) return false;//Unexpected end of file found
	parser.parseline(bufferpointer);
	if (!parser.getlineelement(0,&member,&length)) return false;
	if (strcmp(member,"Finish")==0) return false; //Unexpected end of function found
	if (strcmp(member,"Int")!=0) return true; //Not an Int
	if (!parser.getlineelement(1,&member,&length)) return false;
	sscanf(member," %i",loadedint);
	return true;
}


bool TransXFunction::loadvector(FILEHANDLE scn, VECTOR3 *loadedvector)
{
	char *bufferpointer,*member;
	int length;
	bool ok=oapiReadScenario_nextline(scn,bufferpointer);
	if (!ok) return false;//Unexpected end of file found
	parser.parseline(bufferpointer);
	if (!parser.getlineelement(0,&member,&length)) return false;
	if (strcmp(member,"Finish")==0) return false; //Unexpected end of function found
	if (strcmp(member,"Vector")!=0) return true; //Not a vector
	if (!parser.getlineelement(1,&member,&length)) return false;
	(*loadedvector).x=atof(member);
	if (!parser.getlineelement(2,&member,&length)) return false;
	(*loadedvector).y=atof(member);
	if (!parser.getlineelement(3,&member,&length)) return false;
	(*loadedvector).z=atof(member);
	return true;
}


bool TransXFunction::loadorbit(FILEHANDLE scn,ORBIT *loadorbit)
{
	char *bufferpointer, *member;
	bool ok=oapiReadScenario_nextline(scn,bufferpointer);
	int length;
	if (!ok) return false;
	parser.parseline(bufferpointer);
	if (!parser.getlineelement(0,&member,&length)) return false;//No fields
	if (strcmp(member,"Orbit")!=0) return false; //Not an orbit
	if (!parser.getlineelement(1,&member,&length)) return false;//No second field
	if (strcmp(member,"False")==0)
	{
		//Uninitialised orbit structure
		loadorbit->setinvalid();
		return true;
	}
	VECTOR3 tpos,tvel;
	double gmplanet, timestamp;
	if (!loadvector(scn,&tpos)) return false;
	if (!loadvector(scn,&tvel)) return false;
	if (!loaddouble(scn,&gmplanet)) return false;
	if (!loaddouble(scn,&timestamp)) return false;
	timestamp=(timestamp-simstartMJD)*86400;
	loadorbit->init(tpos,tvel,timestamp,gmplanet);
	return true;
}

void TransXFunction::savehandle(FILEHANDLE scn, OBJHANDLE handle)
{
	char namebuffer[30];
	if (handle!=NULL)
	{
		oapiGetObjectName(handle,namebuffer,30);
	}
	else
	{
		strcpy(namebuffer,"NULL");
	}
	oapiWriteScenario_string(scn,"Handle",namebuffer);
	return;
}


void TransXFunction::saveorbit(FILEHANDLE scn, const ORBIT &saveorbit)
{
	char validvalue[6];
	if (saveorbit.isvalid())
	{
		strcpy(validvalue,"True");
	}
	else
	{
		strcpy(validvalue,"False");
	}
	oapiWriteScenario_string(scn,"Orbit",validvalue);
	if (!saveorbit.isvalid()) return;
	VECTOR3 tpos,tvel;
	saveorbit.getcurrentvectors(&tpos,&tvel);
	savevector(scn,tpos);
	savevector(scn,tvel);
	double planet=saveorbit.getgmplanet();
	savedouble(scn,planet);
	double time=saveorbit.gettimestamp()/86400+simstartMJD;
	savedouble(scn,time);
}






void TransXFunction::setnumberviews(int tnumberviews)
{
	numberviews=tnumberviews;
}

MFDvarhandler *TransXFunction::getvariablehandler()
{
	return &vars;
}

MFDvariable *TransXFunction::getcurrentvariable(int view)
{
	return vars.getcurrent(view);
}


void TransXFunction::sethmajor(OBJHANDLE handle)
{
	hmajor=handle;
	hminor=hmajtarget=NULL;
}

bool TransXFunction::sethminor(OBJHANDLE handle)
{//Virtual to give local ability to tweak this function
	return sethminorstd(handle);
}

bool TransXFunction::sethminorstd(OBJHANDLE handle)
{//Non-virtual function
	if (hmajor==handle) return false;
	if (handle==hmajtarget) hmajtarget=NULL;
	hminor=handle;
	if (hminor==NULL) return false;
	gravbodyratio=pow(oapiGetMass(hminor)/oapiGetMass(hmajor), double (0.8));
	return true;
}

bool TransXFunction::sethmajtarget(OBJHANDLE handle)
{
	if (hmajor==handle) return false;
	if (hminor==handle) hminor=NULL;
	hmajtarget=handle;
	return true;
}

void TransXFunction::sethcraft(OBJHANDLE handle)
{
	hcraft=handle;
}

void TransXFunction::sethbase(OBJHANDLE handle)
{
	hbase=handle;
}

void TransXFunction::sethandles(OBJHANDLE thmajor, OBJHANDLE thminor, OBJHANDLE thtarget, OBJHANDLE thcraft, OBJHANDLE thbase)
{
	hmajor=thmajor;
	gravbodyratio=0;
	sethminor(thminor);
	hmajtarget=thtarget;
	hcraft=thcraft;
	hbase=thbase;
}


void TransXFunction::gethandles(OBJHANDLE *thmajor, OBJHANDLE *thminor, OBJHANDLE *thtarget, OBJHANDLE *thcraft, OBJHANDLE *thbase)
{
	*thmajor=hmajor;
	*thminor=hminor;
	*thtarget=hmajtarget;
	*thcraft=hcraft;
	*thbase=hbase;
}

/*HPEN TransXFunction::SelectDefaultPen(HDC hDC, int value)
{
	return state->GetMFDpointer()->SelectDefaultPen(hDC,value);
}*/




void TransXFunction::initpens(void)								//(rbd+)
{
	pens[0] = CreatePen(PS_SOLID, 0 , RGB(0x00, 0xFF, 0x00));// Green - stands for craft
	pens[1] = CreatePen(PS_SOLID, 0 , RGB(0x00, 0x00, 0xCD));//Blue - stands for planet
	pens[2] = CreatePen(PS_DOT, 0 , RGB(0xCD, 0xCD, 0x00));//Bright yellow - hypos
	pens[3] = CreatePen(PS_SOLID, 0 , RGB(0xFF, 0x00, 0x00));//Bright red - unused, but danger
	pens[4] = CreatePen(PS_SOLID, 0 , RGB(0xC0, 0xC0, 0xC0));//Light Grey
	pens[5] = CreatePen(PS_SOLID, 0 , RGB(0xFF, 0xFF, 0xFF));//Bright white - unused
	//pens[0] = CreatePen(PS_SOLID, 0, RGB(0xCD, 0xCD, 0x00));	// 6 = Solid, Bright Yellow
	//pens[1] = CreatePen(PS_SOLID, 0, RGB(0x00, 0x00, 0xCD));	// 7 = Solid, Medium Blue
	//pens[2] = CreatePen(PS_SOLID, 0, RGB(0xFF, 0xFF, 0xFF));	// 8 = Solid, Bright White
	//pens[3] = CreatePen(PS_SOLID, 0, RGB(0x80, 0x00, 0x00));	// 9 = Solid, Dark Red
}


void TransXFunction::deletepens()
{
	for (int a=0;a<6;a++)
	{
		DeleteObject(pens[a]);
	}
}
															//(rbd-)
HPEN TransXFunction::SelectDefaultPen(HDC hDC, int value)
{
	if(value <6)
	{//(rbd+)
		return((HPEN)SelectObject(hDC, pens[value-1]));		// Custom pen
	}
	else
	{//(rbd-)
		return state->GetMFDpointer()->SelectDefaultPen(hDC,value);
	}
}






void TransXFunction::sethelp(char *help1,char *help2,char *help3,char *help4,char *help5)
{
	strcpy(helpstring1,help1);
	strcpy(helpstring2,help2);
	strcpy(helpstring3,help3);
	strcpy(helpstring4,help4);
	strcpy(helpstring5,help5);
}

void TransXFunction::gethelp(char *help1,char *help2,char *help3,char *help4,char *help5) const
{
	strcpy(help1,helpstring1);
	strcpy(help2,helpstring2);
	strcpy(help3,helpstring3);
	strcpy(help4,helpstring4);
	strcpy(help5,helpstring5);
}

